﻿using System;
using System.Collections.Generic;
using System.IO.IsolatedStorage;
using System.Xml.Linq;
using System.Text.RegularExpressions;
using XElementHelpers;

namespace PhoneAppLLS
{
    public class DataAccessLayer
    {
        public const string FileCurrentVersion = "2.4";

        /// <summary>
        /// This is the version number of the XML file that has been loaded
        /// </summary>
        public string FileVersion
        {
            set;
            get;
        }

        /// <summary>
        /// When true the vault has been loaded
        /// </summary>
        public bool VaultExists
        {
            set;
            get;
        }

        /// <summary>
        /// When true the app is being reactivated
        /// </summary>
        public bool ResumeFromDeactivation
        {
            set;
            get;
        }

        // this is the max number of entries the trail version will allow
        public const int IsTrialLimit = 5;

        /// <summary>
        /// When true the application has been installed as a trial app.
        /// </summary>
        public bool IsTrial
        {
            set;
            get;
        }

        /// <summary>
        /// When true a key has changed in one of the vault entries and it should be
        /// reloaded
        /// </summary>
        public bool ListReloadRequired
        {
            set;
            get;
        }

        /// <summary>
        /// When true show all the information, when false just the title
        /// </summary>
        public bool ListShowAll
        {
            set;
            get;
        }

        /// <summary>
        /// Dictionary containing all the current vault entries
        /// </summary>
        public Dictionary<string, VaultEntry> VaultEntries
        {
            set;
            get;
        }

        /// <summary>
        /// Constructor
        /// </summary>
        public DataAccessLayer()
        {
            Reset();
        }

        /// <summary>
        /// Reset all the Vault settings to a default value
        /// </summary>
        public void Reset()
        {
            ResumeFromDeactivation = false;

            VaultEntries = new Dictionary<string, VaultEntry>();
            VaultExists = false;

            // load any pre-defined settings that may exist (service URI etc)
            LoadSettingsFromIsolatedStorage();
        }

        /// <summary>
        /// If there is a MasterKey entry in IS then the vault exists
        /// </summary>
        /// <returns></returns>
        public bool TestVaultExists()
        {

            string vault = GetSetting("MasterKey");
            VaultExists = (vault == "") ? false : true;

            return VaultExists;
        }

        /// <summary>
        /// Pick up default values from storage
        /// </summary>
        public void LoadSettingsFromIsolatedStorage()
        {
            try
            {
            }
            catch
            {
                throw;
            }
        }

        /// <summary>
        /// Load the Vault from isolated storage
        /// 
        /// This method has to also deal with a v1.4 Vault that may already be in isolated storage
        /// </summary>
        public void LoadVaultFromIsolatedStorage()
        {
            try
            {
                VaultEntries = new Dictionary<string, VaultEntry>();

                string vault = GetSetting("MasterKey");
                if (vault != "")
                {
                    XElement xVault = XElement.Parse(vault);
                    LoadVaultFromXML(xVault);

                    VaultExists = true;
                }
                else
                {
                    VaultExists = false;
                }
            }
            catch
            {
                throw;
            }
        }

        /// <summary>
        /// Load versioned XML 
        /// </summary>
        /// <param name="Vault"></param>
        /// <returns></returns>
        public bool LoadVaultFromXML(XElement Vault)
        {
            try
            {
                VaultEntries = new Dictionary<string, VaultEntry>();

                // get the vault version number, anything other than a value is version 1.4
                FileVersion = Vault.Element("FileVersion").ToStringNull();

                foreach (XElement xEntry in Vault.Element("Vault").Elements("Entry"))
                {
                    VaultEntry entry = new VaultEntry();
                    entry.LoadEntryXML(xEntry);
                    Add(entry);
                }

                SaveVaultToIsolatedStorage();

                // a valid vault has been loaded
                VaultExists = true;
            }
            catch
            {
                VaultExists = false;
                throw;
            }

            return true;
        }

        /// <summary>
        /// Fetch the named value from isolated storage.
        /// Will return an empty string if the entry does not exist
        /// </summary>
        /// <param name="Name"></param>
        /// <returns></returns>
        private string GetSetting(string Name)
        {
            IsolatedStorageSettings settings = IsolatedStorageSettings.ApplicationSettings;

            string value = "";
            settings.TryGetValue(Name, out value);

            if (value == null) value = "";

            return value;
        }

        /// <summary>
        /// Save the Vault to isolated storage
        /// 
        /// The Vault can only be saved if the password that was entered was valid.  If it is saved
        /// with an invalid password then it will of course be corrupted
        /// </summary>
        public void SaveVaultToIsolatedStorage()
        {
            IsolatedStorageSettings settings = IsolatedStorageSettings.ApplicationSettings;
            settings.Clear();

            XElement vault = SaveVaultToXML();
            settings["MasterKey"] = vault.ToString();

            settings.Save();
        }

        /// <summary>
        /// Save the current dictionary to an XML doc
        /// </summary>
        /// <returns></returns>
        public XElement SaveVaultToXML()
        {
            XElement xFile = new XElement("MasterKey");

            // this is the XML file format version
            xFile.Add(new XElement("FileVersion", FileCurrentVersion));

            XElement xEntries = new XElement("Vault");
            xFile.Add(xEntries);

            try
            {
                int c = 1;
                foreach (VaultEntry entry in VaultEntries.Values)
                {
                    xEntries.Add(entry.SaveEntryXML());
                    c++;
                }
            }
            catch
            {
                throw;
            }

            return xFile;
        }

        /// <summary>
        /// Add the entry to the dictionary
        ///     - if the entry does not exist then add it
        ///     - if the entry exists then update the rest of the info
        /// </summary>
        /// <param name="Entry"></param>
        /// <returns>true if a change has been made</returns>
        public bool Add(VaultEntry Entry)
        {
            try
            {
                if (Entry.Caption == "")
                {
                    throw new Exception("caption cannot be empty");
                }

                // encrypt the entry on the way into the dictionary
                VaultEntry entry = new VaultEntry();
                VaultEntries.TryGetValue(Entry.Caption, out entry);

                if (entry == null)
                {
                    // entry does not exist so simply add it (this will mean that a change to the caption will result in two entries)
                    VaultEntries.Add(Entry.Caption, Entry);

                    // sort the entries
                    SortEntries();

                    // the dictionary has changed
                    ListReloadRequired = true;
                }
                else
                {
                    // entry exists so simply update it
                    if (entry.Password == Entry.Password && entry.Identity == Entry.Identity && entry.OtherInfo == Entry.OtherInfo)
                    {
                        // no change so go back
                        return false;
                    }
                    else
                    {
                        // update the info
                        entry.Password = Entry.Password;
                        entry.Identity = Entry.Identity;
                        entry.OtherInfo = Entry.OtherInfo;
                    }
                }
            }
            catch
            {
                throw;
            }

            return true;
        }

        /// <summary>
        /// As the Silverlight Dictionary does not have a sorted option its necessary to outsource it...
        /// 1. Copy the keys to a string array
        /// 2. Assign the array to a List
        /// 3. Sort the List
        /// 4. Create a new ValueEntry dictionary based on the key order in the List
        /// 5. Assign this dictionary to the existing ValueEntries dictionary
        /// 
        /// If there is an error the original dictionary will be retained.
        /// </summary>
        public void SortEntries()
        {
            try
            {
                // sort the keys - Silverlight Dictionary does not have a sort option :-(
                string[] array = new string[VaultEntries.Keys.Count];
                VaultEntries.Keys.CopyTo(array, 0);
                List<string> list = new List<string>(array);
                list.Sort();

                // now recreate the dictioanry                
                Dictionary<string, VaultEntry> sortedEntries = new Dictionary<string, VaultEntry>();
                VaultEntry entry = new VaultEntry();

                foreach (string key in list)
                {
                    VaultEntries.TryGetValue(key, out entry);
                    sortedEntries.Add(key, entry);
                }

                VaultEntries = sortedEntries;
            }
            catch
            {
                throw;
            }
        }

        /// <summary>
        /// Validate that the entry can be added to the dictionary.
        /// </summary>
        /// <param name="Entry"></param>
        /// <returns>true if the entry is valid</returns>
        public bool Validate(VaultEntry Entry)
        {
            try
            {
                if (Entry.Caption == "")
                {
                    throw new Exception("Caption cannot be empty");
                }

                return true;
            }
            catch
            {
                throw;
            }
        }

        /// <summary>
        /// Delete the entry from the dictionary.
        /// If the Key cannot be found an error is thrown
        /// If the Key is found the entry is deleted
        /// </summary>
        /// <param name="Entry"></param>
        /// <returns></returns>
        public void Delete(string Key)
        {
            try
            {
                VaultEntry entry = new VaultEntry();
                VaultEntries.TryGetValue(Key, out entry);

                if (entry == null) throw new Exception(string.Format("Cannot find: {0}", Key));

                VaultEntries.Remove(Key);

                ListReloadRequired = true;
            }
            catch
            {
                throw;
            }
        }

        /// <summary>
        /// Get the item from dictionary. Returns either null or a clone of the entry
        /// </summary>
        /// <param name="Index"></param>
        /// <returns></returns>
        public VaultEntry Get(string Key)
        {
            try
            {
                VaultEntry entry = new VaultEntry();
                VaultEntries.TryGetValue(Key, out entry);

                if (entry == null) return null;

                return entry.Clone();
            }
            catch
            {
                throw;
            }
        }

        /// <summary>
        /// Get the item from dictionary. Returns either null or a clone of the entry
        /// </summary>
        /// <param name="Index"></param>
        /// <returns></returns>
        public VaultEntry Get(int Index)
        {
            try
            {
                VaultEntry[] array = new VaultEntry[VaultEntries.Count];
                VaultEntries.Values.CopyTo(array, 0);

                if (Index >= VaultEntries.Count || Index < 0) return null;

                return array[Index].Clone();
            }
            catch
            {
                throw;
            }
        }

        /// <summary>
        /// Create a Test Vault
        /// </summary>
        /// <param name="Password"></param>
        /// <param name="Salt"></param>
        public void TestVault(string Password, string Salt, bool Trail)
        {
            if (!VaultExists)
            {
                // debug - add some coded examples for the system to use when working with the Emulator
                VaultExists = true;
                this.IsTrial = Trail;

                Add(new VaultEntry("My Bank", "Joe Bloggs", "4567", ""));
                Add(new VaultEntry("Tesco Online", "jblogs@msn.com", "TescoOnline23", ""));
                Add(new VaultEntry("Keysafe", "", "2389a", ""));
                Add(new VaultEntry("SharePoint Admin on IServer7", "emeaMOSSadmin", "blagh2", ""));
                if (Trail == false)
                {
                    for (int c = 0; c < 20; c++)
                    {
                        Add(new VaultEntry(string.Format("Test {0}", c), "", "blagh2", ""));
                    }

                    for (int c = 0; c < 5; c++)
                    {
                        Add(new VaultEntry(string.Format("Windows {0}", c), "", "blagh2", ""));
                    }
                    //Add(new VaultEntry("Windows", "rolle polle", "blagh2", "this is some text"));
                    Add(new VaultEntry("Batter", "2 eggs n flour", "blagh2", "beat together"));
                    Add(new VaultEntry("Butter", "Dorset pride", "butter", "makes a good cake"));
                }

                // save the vault
                SaveVaultToIsolatedStorage();

                // reset the vault and then reload it...
                Reset();

                // and reload it...
                LoadVaultFromIsolatedStorage();
            }
        }
    }
}
